@@ -3,6 +3,9 @@
 todo:
 
 to remove ambiguity regarding patch_x,patch_y,window_x,window_y
+extend soft_segmentation to n_patterns > 2 cases.
+store descriptors, and avoid double calculating the same image
+
 """
 
 
@@ -15,10 +18,11 @@ from stemrotational_symmetry_descriptors import get_rotational_symmetry_descript
 from stemreflection_symmetry_descriptors import get_reflection_symmetry_descriptors
 from sklearn.cluster import KMeans
 from scipy.ndimage import map_coordinates
+from scipy.interpolate import NearestNDInterpolator
 import gc
 
 descriptors_implemented = ['power_spectrum','local_correlation_map','rotational_symmetry_maximums','reflection_symmetry_maximums']
-
+methods_implemented = ['direct','fft']
 class segmentationSTEM:
     def __init__(self,n_patterns=2, 
                  window_x=21,window_y=21,
@@ -27,6 +31,7 @@ class segmentationSTEM:
                  upsampling=True,
                  pca_fitted=None,kmeans_init_centers=None,
                  one_step_kmeans=False,
+                 removing_mean=True,
                  # 
                  num_reflection_plane=10,
                  # parameters associated with rotational with rotational_symmetry_maximums descriptors
@@ -35,6 +40,8 @@ class segmentationSTEM:
                  patch_x=20,patch_y=20, max_num_points=100, parallel=True,
                  #
                  power_spectrum_logarithm=True,
+                 soft_segmentation=False,
+                 method = 'direct',
                  ):
         
         """
@@ -59,12 +66,15 @@ class segmentationSTEM:
         self._segmentation_labels = []
         if descriptor_name not in descriptors_implemented:
             raise ValueError('descriptor_name should be in ', descriptors_implemented)
+        if method not in methods_implemented:
+            raise ValueError('method should be in', methods_implemented)
         self.paras = {'n_patterns':n_patterns,
                       'patch_x':patch_x,
                       'patch_y':patch_y,
                       'window_x':window_x,
                       'window_y':window_y,
                       'descriptor_name':descriptor_name,
+                      'removing_mean': removing_mean,
                       'radius':radius,
                       'nr':nr,
                       'nt':nt,
@@ -78,9 +88,12 @@ class segmentationSTEM:
                       'n_PCA_components':n_PCA_components,
                       'max_num_points':max_num_points, 
                       'power_spectrum_logarithm':power_spectrum_logarithm,
+                      'soft_segmentation':soft_segmentation,
+                      'method':method,
                       'parallel':parallel}
 
-    def get_PCA_components(self, image):
+    
+    def get_descriptors(self,image):    
         self.check_image_validity(image)
         if self.paras['descriptor_name'] is 'local_correlation_map':
             descriptors = get_descriptor(image,self.paras['patch_x'],
@@ -89,9 +102,11 @@ class segmentationSTEM:
                                         self.paras['window_y'],
                                         self.paras['max_num_points'],
                                         step = self.paras['step'],
-                                        parallel=self.paras['parallel'])
+                                        parallel=self.paras['parallel'],
+                                        method=self.paras['method'],
+                                        removing_mean=self.paras['removing_mean'])
         elif self.paras['descriptor_name'] is 'power_spectrum':
-            descriptors = get_power_spectrum_m1(image,2*self.paras['window_x'],2*self.paras['window_y'], step=self.paras['step'],logarithm=self.paras['power_spectrum_logarithm'])
+            descriptors = get_power_spectrum_m1(image,self.paras['window_x'],self.paras['window_y'], step=self.paras['step'],logarithm=self.paras['power_spectrum_logarithm'])
         elif self.paras['descriptor_name'] is 'rotational_symmetry_maximums':
             descriptors = get_rotational_symmetry_descriptors(image, window_x=self.paras['window_x'], window_y=self.paras['window_y'],
                                                               radius=self.paras['radius'],nr=self.paras['nr'], nt=self.paras['nt'],
@@ -101,7 +116,12 @@ class segmentationSTEM:
                                                               radius=self.paras['radius'],nr=self.paras['nr'], nt=self.paras['nt'],
                                                               num_reflection_plane=self.paras['num_reflection_plane'],
                                                               step_symmetry_analysis=self.paras['step'],num_max=self.paras['num_max'])
-            
+        self._descriptors = descriptors 
+        return self._descriptors
+
+    def get_PCA_components(self, image):
+        
+        descriptors = self.get_descriptors( image)
         n_components = self.paras['n_PCA_components']
         shape = descriptors.shape
         if self.paras['pca_fitted'] is None:
@@ -123,7 +143,7 @@ class segmentationSTEM:
         diff = shape_image[0] - len(x_index)
         diff_left = int( diff / 2.)
         diff_right = diff - diff_left
-        left_index = np.flip(-np.arange(1, 1+diff_left))
+        left_index = np.flip(-np.arange(1, 1+diff_left),axis=0)
         right_index = len(x_index)+np.arange(diff_right)
         x_index = np.concatenate((left_index, x_index, right_index))
 
@@ -131,7 +151,7 @@ class segmentationSTEM:
         diff = shape_image[1] - len(y_index)
         diff_left = int( diff / 2.)
         diff_right = diff - diff_left
-        left_index = np.flip(-np.arange(1, 1+diff_left))
+        left_index = np.flip(-np.arange(1, 1+diff_left),axis=0)
         right_index = len(y_index)+np.arange(diff_right)
         y_index = np.concatenate((left_index, y_index, right_index))
             
@@ -139,17 +159,46 @@ class segmentationSTEM:
         x_grid = (x_grid / step).flatten()
         y_grid = (y_grid / step).flatten()
         coords = np.vstack((x_grid, y_grid))
-        labels_up = map_coordinates(labels, coords,mode='nearest')
-        labels_up = np.reshape(labels_up, (shape_image[1], shape_image[0])).T
-        labels_up = np.round(labels_up).astype(np.int32) % self.paras['n_patterns']
+        if self.paras['soft_segmentation'] is False:
+            x_labels, y_labels = np.meshgrid(np.arange(shape_labels[0]),np.arange(shape_labels[1]),indexing='ij')
+            x_labels = x_labels.flatten()
+            y_labels = y_labels.flatten()
+            input_coords = np.vstack([x_labels, y_labels]).T
+            interpolator = NearestNDInterpolator(input_coords, labels.flatten())
+            labels_up = interpolator(coords.T)
+            labels_up = np.reshape(labels_up, shape_image).T
+            #labels_up = np.round(labels_up).astype(np.int32) % self.paras['n_patterns']
+        else:
+            labels_up = map_coordinates(labels, coords,mode='nearest')
+            labels_up = np.reshape(labels_up, (shape_image[1], shape_image[0])).T
+            labels_up = np.clip(labels_up,0.,self.paras['n_patterns']-1.)
+ 
         return labels_up
 
+    def perform_soft_segmentation(self, cluster_centers,features):
+        """
+        cluster_centers..........ndarray of shape(n_clusters, n_features)
+        features.................ndarray of shape(n_samples, n_features)
+        """
+        import scipy
+        (n_clusters, n_features) = cluster_centers.shape
+        (n_samples, n_features_t) = features.shape
+        if not (n_clusters == 2):
+            raise ValueError('currently only implemented for the two-clusters case')
+        if not (n_features == n_features_t):
+            raise ValueError('# of features not equal')
+        distances = np.zeros((n_samples, n_clusters))
+        for i in range(n_clusters):
+            distances[:,i] = np.linalg.norm(features - cluster_centers[i:i+1,:],axis=1)
+        return scipy.special.softmax(distances,axis=1)[:,0]
+
 
     def perform_clustering(self, image):
         """
         image..............2D numpy array
         upsampling.........Boolean, if True, upsampling in order to match the shape of image.
         """
+        image = image.astype(np.float32)
         features = self.get_PCA_components(image)
         shape = features.shape
         if self.paras['one_step_kmeans'] is True:
@@ -162,7 +211,12 @@ class segmentationSTEM:
             kmeans = KMeans(n_clusters=self.paras['n_patterns'],max_iter=max_iter,init=self.paras['kmeans_init_centers'])
         kmeans.fit(np.reshape(features,(-1,shape[2])))
         self._kmeans = kmeans
-        labels = np.reshape(kmeans.labels_, (shape[0],shape[1]))
+        if self.paras['soft_segmentation'] is False:
+            labels = np.reshape(kmeans.labels_, (shape[0],shape[1]))
+        else:
+            soft_labels = self.perform_soft_segmentation(kmeans.cluster_centers_, np.reshape(features,(-1,shape[2])))
+            labels = np.reshape(soft_labels, (shape[0],shape[1]))
+        # this line looks useless
         self._segmentation_labels = np.zeros_like(image,dtype=np.int32) - 1
         shape_image = image.shape
         window_x = self.paras['window_x']
@@ -193,3 +247,4 @@ class segmentationSTEM:
             raise ValueError("image should be 2D numpy array")
 
 
+
